from datetime import datetime

import httpx
import respx

from wallaroo.assay_config import *
import json
import pytest
import unittest
import responses

import wallaroo
from wallaroo.pipeline import Pipeline
from . import testutil
from .reusable_responders import add_default_workspace_responder, add_get_topic_name_responder


class TestAssayConfig(unittest.TestCase):
    def setUp(self):
        self.gql_client = testutil.new_gql_client(endpoint="http://api-lb:8080/v1/graphql")
        self.test_client = wallaroo.Client(
            gql_client=self.gql_client,
            auth_type="test_auth",
            api_endpoint="http://api-lb:8080",
            config={"default_arch": "x86"},
            request_timeout=2,  # this field used only by deployment.wait_for_running() tests
        )

    def test_ensure_tz(self):
        now = datetime.now()
        nowutz = ensure_tz(now)
        assert now != nowutz

        now = datetime.now(tz=timezone.utc)
        nowutz = ensure_tz(now)
        assert now == nowutz

    def test_calc_bins(self):
        assert calc_bins(100_000, "auto") == "auto"
        assert calc_bins(100_000, None) == 50
        assert calc_bins(650, None) == 25
        assert calc_bins(660, None) == 25

    def test_unwrap(self):
        v = unwrap(3)
        assert v == 3

        with pytest.raises(Exception):
            v = unwrap(None)

    @responses.activate
    def test_fixed_baseline(self):
        add_default_workspace_responder()
        pipeline_name = "mypipeline"
        model_name = "mymodel"
        bl = FixedBaseline(pipeline_name, model_name, datetime.utcnow(), datetime.now())
        json_dict = json.loads(bl.to_json())
        fixed = json_dict["calculated"]["fixed_window"]
        assert fixed["pipeline"] == pipeline_name
        assert fixed["model"] == model_name

    @respx.mock(assert_all_called=True)
    def test_fixed_baseline_builder(self, respx_mock):
        pipeline_name = "mypipeline"
        model_name = "mymodel"
        resp = respx_mock.post(f"{self.test_client.api_endpoint}/v1/api/assays/summarize", ).mock(
            return_value=httpx.Response(200, json={
                "count": 188,
                "min": 11.986584663391112,
                "max": 14.29722023010254,
                "mean": 13.031112508570894,
                "median": 12.956134796142578,
                "std": 0.4770556767131347,
                "edges": [11.986584663391112, 12.622478485107422, 12.854415893554688, 13.064453125, 13.440485000610352,
                          14.29722023010254, None],
                "edge_names": ["left_outlier", "q_20", "q_40", "q_60", "q_80", "q_100", "right_outlier"],
                "aggregated_values": [0.0, 0.20212765957446807, 0.19680851063829788, 0.20212765957446807,
                                      0.19680851063829788, 0.20212765957446807, 0.0],
                "aggregation": "Density",
                "start": "2023-01-01T00:00:00+00:00",
                "end": "2023-01-02T00:00:00+00:00",
            })
        )
        fbb = FixedWindowBaselineBuilder(client=self.test_client, pipeline_name=pipeline_name)
        sc = UnivariateContinousSummarizerConfig(
            BinMode.QUANTILE,
            Aggregation.DENSITY,
            Metric.SUMDIFF,
            10,
            None,
            None,
            None,
            True,
        )
        fbb = fbb.add_summarizer(sc)
        jd = json.loads(fbb.to_json())
        assert jd["pipeline_name"] == pipeline_name
        assert jd["model_name"] is None
        assert jd["start"] is None
        assert jd["end"] is None

        with pytest.raises(Exception):
            jd.build()

        fbb.add_model_name(model_name)
        fbb.add_start(datetime.now())
        fbb.add_end(datetime.now())

        fb = fbb.build()
        assert fb is not None
        self.assertIsInstance(fb, StaticBaseline)
        self.assertEqual(fb.static["count"], 188)

    def test_summarizer_config(self):
        ts = SummarizerConfig()
        assert ts is not None

    def test_uc_summarizer_config(self):
        num_bins = 10
        ucs = UnivariateContinousSummarizerConfig(
            BinMode.QUANTILE,
            Aggregation.DENSITY,
            Metric.SUMDIFF,
            num_bins,
            None,
            None,
            None,
            True,
        )
        assert ucs is not None
        ucs_dict = json.loads(ucs.to_json())
        assert ucs_dict["num_bins"] == num_bins

    def test_uc_sum_builder(self):
        ucsb = UnivariateContinousSummarizerBuilder()
        assert ucsb is not None

        ucsb.add_aggregation(Aggregation.EDGES)
        ucsb.add_metric(Metric.MAXDIFF)

        sum_config = ucsb.build()
        sum = json.loads(sum_config.to_json())
        assert sum["num_bins"] == 5
        assert sum["type"] == "UnivariateContinuous"
        assert sum["aggregation"] == "Edges"
        assert sum["metric"] == "MaxDiff"

    def test_window_config(self):
        wc = WindowConfig(pipeline_name="pipeline_name", model_name="model_name", width="3 hour")
        wd = json.loads(wc.to_json())
        assert wd["model_name"] == "model_name"

    def test_window_builder(self):
        pipeline_name = "mypipeline"
        model_name = "mymodel"
        wb = WindowBuilder(pipeline_name)
        wb.add_model_name(model_name)
        wb.add_width(hours=3)
        assert wb is not None
        window_dict = json.loads(wb.build().to_json())
        assert window_dict["model_name"] == model_name
        assert window_dict["width"] == "3 hours"

    def test_config_encoder(self):
        d = datetime.now()
        o = {"date": d}
        with pytest.raises(Exception):
            s = json.dumps(o)

        s = json.dumps(o, default=ConfigEncoder)
        assert s is not None

    def test_assay_config(self):
        bl = FixedBaseline("pipeline_name", "model_name", datetime.utcnow(), datetime.now())
        wc = WindowConfig("pipeline_name", "model_name", "3 hour")
        sc = UnivariateContinousSummarizerConfig(
            BinMode.QUANTILE,
            Aggregation.DENSITY,
            Metric.SUMDIFF,
            10,
            None,
            None,
            None,
            True,
        )

        ac = AssayConfig(
            None,
            "test",
            0,
            "test",
            True,
            "test",
            bl,
            wc,
            sc,
            None,
            0.5,
            datetime.now(),
            None,
        )
        ad = json.loads(ac.to_json())
        print(ac.to_json())
        assert ad["name"] == "test"

    @responses.activate
    @respx.mock(assert_all_mocked=True)
    def test_assay_builder(self, respx_mock):
        add_default_workspace_responder()
        resp = respx_mock.post(f"{self.test_client.api_endpoint}/v1/api/assays/summarize", ).mock(
            return_value=httpx.Response(200, json={
                "count": 188,
                "min": 11.986584663391112,
                "max": 14.29722023010254,
                "mean": 13.031112508570894,
                "median": 12.956134796142578,
                "std": 0.4770556767131347,
                "edges": [11.986584663391112, 12.622478485107422, 12.854415893554688, 13.064453125, 13.440485000610352,
                          14.29722023010254, None],
                "edge_names": ["left_outlier", "q_20", "q_40", "q_60", "q_80", "q_100", "right_outlier"],
                "aggregated_values": [0.0, 0.20212765957446807, 0.19680851063829788, 0.20212765957446807,
                                      0.19680851063829788, 0.20212765957446807, 0.0],
                "aggregation": "Density",
                "start": "2023-01-01T00:00:00+00:00",
                "end": "2023-01-02T00:00:00+00:00",
            })
        )
        ab = AssayBuilder(
            client=self.test_client,
            name="n",
            pipeline_id=0,
            pipeline_name="test",
            model_name="test",
            baseline_start=datetime.now(),
            baseline_end=datetime.now(),
            iopath="inputs 0 0",
        )
        assert ab.name == "n"
        ab.add_name("ab")

        with pytest.raises(Exception):
            ab.add_iopath("foo")

        ab.add_iopath(" inputs 0 0 ")

        ad = ab.build()
        self.assertEqual(ad.name, "ab")
        self.assertIsInstance(ad.baseline, StaticBaseline)
        self.assertEqual(ad.baseline.static["count"], 188)

    @responses.activate
    @respx.mock(assert_all_mocked=True)
    def test_assay_builder_window_width(self, respx_mock):
        add_default_workspace_responder()
        resp = respx_mock.post(f"{self.test_client.api_endpoint}/v1/api/assays/summarize", ).mock(
            return_value=httpx.Response(200, json={
                "count": 188,
                "min": 11.986584663391112,
                "max": 14.29722023010254,
                "mean": 13.031112508570894,
                "median": 12.956134796142578,
                "std": 0.4770556767131347,
                "edges": [11.986584663391112, 12.622478485107422, 12.854415893554688, 13.064453125, 13.440485000610352,
                          14.29722023010254, None],
                "edge_names": ["left_outlier", "q_20", "q_40", "q_60", "q_80", "q_100", "right_outlier"],
                "aggregated_values": [0.0, 0.20212765957446807, 0.19680851063829788, 0.20212765957446807,
                                      0.19680851063829788, 0.20212765957446807, 0.0],
                "aggregation": "Density",
                "start": "2023-01-01T00:00:00+00:00",
                "end": "2023-01-02T00:00:00+00:00",
            })
        )
        assay_builder = AssayBuilder(
            client=self.test_client,
            name="n",
            pipeline_id=0,
            pipeline_name="test",
            model_name="test",
            baseline_start=datetime.now(),
            baseline_end=datetime.now(),
            iopath="inputs 0 0",
        )
        assay_builder.window_builder().add_width(hours=12)

        acd = assay_builder.build()
        self.assertEqual(acd.window.width, "12 hours")
        self.assertIsNone(acd.window.start)
        self.assertIsNone(acd.window.interval)

        assay_builder.window_builder().add_interval(hours=4)
        window_start_date = datetime.now()
        assay_builder.window_builder().add_start(window_start_date)
        acd = assay_builder.build()
        self.assertEqual(acd.window.width, "12 hours")
        self.assertEqual(acd.window.interval, "4 hours")
        self.assertEqual(acd.window.start, window_start_date.astimezone(tz=timezone.utc))

        # Invalid interval type
        with pytest.raises(Exception):
            assay_builder.window_builder().add_interval(fortnights=2)

        # Multiple intervals invalid
        with pytest.raises(Exception):
            assay_builder.window_builder().add_interval(weeks=1, hours=4)

    @responses.activate
    @respx.mock(assert_all_mocked=True)
    def test_assay_builder_bin_settings(self, respx_mock):
        add_default_workspace_responder()
        weights = [1.0] * 7
        resp = respx_mock.post(f"{self.test_client.api_endpoint}/v1/api/assays/summarize", ).mock(
            return_value=httpx.Response(200, json={
                "count": 188,
                "min": 11.986584663391112,
                "max": 14.29722023010254,
                "mean": 13.031112508570894,
                "median": 12.956134796142578,
                "std": 0.4770556767131347,
                "edges": [11.986584663391112, 12.622478485107422, 12.854415893554688, 13.064453125, 13.440485000610352,
                          14.29722023010254, None],
                "edge_names": ["left_outlier", "q_20", "q_40", "q_60", "q_80", "q_100", "right_outlier"],
                "aggregated_values": [0.0, 0.20212765957446807, 0.19680851063829788, 0.20212765957446807,
                                      0.19680851063829788, 0.20212765957446807, 0.0],
                "aggregation": "Density",
                "start": "2023-01-01T00:00:00+00:00",
                "end": "2023-01-02T00:00:00+00:00",
            })
        )
        ab = AssayBuilder(
            client=self.test_client,
            name="n",
            pipeline_id=0,
            pipeline_name="test",
            model_name="test",
            baseline_start=datetime.now(),
            baseline_end=datetime.now(),
            iopath="inputs 0 0",
        )
        with pytest.raises(Exception):
            ab.summarizer_builder.add_bin_weights([1])

        ab.summarizer_builder.add_bin_weights(weights)

        ad = ab.build()
        self.assertEqual(ad.summarizer.bin_weights, weights)

        with pytest.raises(Exception):
            ab.summarizer_builder.add_num_bins(7)

        ab.summarizer_builder.add_bin_weights(None)
        ab.summarizer_builder.add_num_bins(7)

        ad = ab.build()
        self.assertIsNone(ad.summarizer.bin_weights)
        self.assertEqual(ad.summarizer.num_bins, 7)

    @responses.activate
    @respx.mock(assert_all_mocked=True)
    def test_assay_builder_add_edges(self, respx_mock):
        add_default_workspace_responder()
        edges = [1.0] * 6
        resp = respx_mock.post(f"{self.test_client.api_endpoint}/v1/api/assays/summarize", ).mock(
            return_value=httpx.Response(200, json={
                "count": 188,
                "min": 11.986584663391112,
                "max": 14.29722023010254,
                "mean": 13.031112508570894,
                "median": 12.956134796142578,
                "std": 0.4770556767131347,
                "edges": [11.986584663391112, 12.622478485107422, 12.854415893554688, 13.064453125, 13.440485000610352,
                          14.29722023010254, None],
                "edge_names": ["left_outlier", "q_20", "q_40", "q_60", "q_80", "q_100", "right_outlier"],
                "aggregated_values": [0.0, 0.20212765957446807, 0.19680851063829788, 0.20212765957446807,
                                      0.19680851063829788, 0.20212765957446807, 0.0],
                "aggregation": "Density",
                "start": "2023-01-01T00:00:00+00:00",
                "end": "2023-01-02T00:00:00+00:00",
            })
        )
        ab = AssayBuilder(
            client=self.test_client,
            name="n",
            pipeline_id=0,
            pipeline_name="test",
            model_name="test",
            baseline_start=datetime.now(),
            baseline_end=datetime.now(),
            iopath="inputs 0 0",
        )
        # too few edges
        with pytest.raises(Exception):
            ab.summarizer_builder.add_bin_edges([1])

        # edges == to number of bins
        ab.summarizer_builder.add_bin_edges(edges[1:])
        # edges specifying left outlier
        ab.summarizer_builder.add_bin_edges(edges)

        # too many edges
        with pytest.raises(Exception):
            ab.summarizer_builder.add_bin_edges([1.0] * 8)

        # check that that we can't build with the wrong bin mode
        with pytest.raises(Exception):
            ad = json.loads(ab.build().to_json())

        # check that the edges get into the json and back
        ab.summarizer_builder.add_bin_mode(BinMode.PROVIDED, edges)
        ad = ab.build()
        self.assertListEqual(ad.summarizer.provided_edges, edges)

        # try to change num bins
        moar_bins = 10
        with pytest.raises(Exception):
            ab.summarizer_builder.add_num_bins(moar_bins)

        # clear out the edges and add more bins
        ab.summarizer_builder.add_bin_mode(BinMode.EQUAL)
        ab.summarizer_builder.add_num_bins(moar_bins)

        ad = ab.build()
        self.assertIsNone(ad.summarizer.provided_edges)
        self.assertEqual(ad.summarizer.num_bins, moar_bins)

        ab.summarizer_builder.add_bin_mode(BinMode.PROVIDED, [1.0] * moar_bins)
        ad = ab.build()
        self.assertEqual(len(ad.summarizer.provided_edges), moar_bins)

    @responses.activate
    def test_assay_builder_with_baseline_data(self):
        add_default_workspace_responder()
        baseline_data = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9])
        ab = AssayBuilder(
            client=self.test_client,
            name="n",
            pipeline_id=0,
            pipeline_name="test",
            model_name="test",
            iopath="output 2 0",
            baseline_data=baseline_data
        )
        assert isinstance(ab.baseline_builder, VectorBaselineBuilder)
        df = ab.baseline_dataframe()
        assert isinstance(df, pd.DataFrame)
        assert df.columns == ["output_2_0"]

    @responses.activate
    @respx.mock(assert_all_mocked=True)
    def test_build_assay_with_static_baseline_data(self, respx_mock):
        add_default_workspace_responder()
        add_get_topic_name_responder()
        resp = respx_mock.post(f"{self.test_client.api_endpoint}/v1/api/assays/summarize", ).mock(
            return_value=httpx.Response(200, json={
                "count": 188,
                "min": 11.986584663391112,
                "max": 14.29722023010254,
                "mean": 13.031112508570894,
                "median": 12.956134796142578,
                "std": 0.4770556767131347,
                "edges": [11.986584663391112, 12.622478485107422, 12.854415893554688, 13.064453125, 13.440485000610352,
                          14.29722023010254, None],
                "edge_names": ["left_outlier", "q_20", "q_40", "q_60", "q_80", "q_100", "right_outlier"],
                "aggregated_values": [0.0, 0.20212765957446807, 0.19680851063829788, 0.20212765957446807,
                                      0.19680851063829788, 0.20212765957446807, 0.0],
                "aggregation": "Density",
                "start": "2023-01-01T00:00:00+00:00",
                "end": "2023-01-02T00:00:00+00:00",
            })
        )
        pipeline = Pipeline(
            client=self.test_client,
            data={
                "id": 1,
                "pipeline_id": "x",
                "created_at": datetime.now().isoformat(),
                "updated_at": datetime.now().isoformat(),
                "pipeline_versions": [{"id": 1}],
                "visibility": "pUbLIC",
            },
        )
        window_builder = WindowBuilder("test").add_model_name("model_name").add_start(datetime.now())

        baseline_data = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9])
        a = self.test_client.build_assay(
            assay_name="test",
            pipeline=cast(Pipeline, pipeline),
            model_name="model_name",
            iopath="output 0 0",
            baseline_data=baseline_data
        )
        a.window_builder_ = window_builder
        test_baseline_assay_config = a.build()
        self.assertIsInstance(test_baseline_assay_config.baseline, StaticBaseline)
        self.assertIsNotNone(test_baseline_assay_config.baseline.static)
        self.assertEqual(test_baseline_assay_config.baseline.static["count"], 188)
        self.assertIsNotNone(test_baseline_assay_config.window.start)
