"""This module features the Baseline class that configures baselines."""

from datetime import datetime
from typing import List, Union, cast

import dateutil

from wallaroo.assay_config import (
    AssayConfig as V1AssayConfig,
    BinMode as V1BinMode,
    CalculatedBaseline as V1CalculatedBaseline,
    FixedBaseline as V1FixedBaseline,
    StaticBaseline as V1StaticBaseline,
    UnivariateContinousSummarizerConfig as V1Summarizer,
)
from wallaroo.unwrap import unwrap
from wallaroo.wallaroo_ml_ops_api_client.models import (
    BaselineType0 as MLOpsSummaryBaseline,
    BaselineType1 as MLOpsStaticBaseline,
)
from wallaroo.wallaroo_ml_ops_api_client.models.aggregation import Aggregation
from wallaroo.wallaroo_ml_ops_api_client.models.bin_mode_type_0 import BinModeType0
from wallaroo.wallaroo_ml_ops_api_client.models.bin_mode_type_1 import BinModeType1
from wallaroo.wallaroo_ml_ops_api_client.models.bin_mode_type_2 import BinModeType2
from wallaroo.wallaroo_ml_ops_api_client.models.bin_mode_type_4 import BinModeType4
from wallaroo.wallaroo_ml_ops_api_client.models.bins import Bins
from wallaroo.wallaroo_ml_ops_api_client.models.field_tagged_summaries import (
    FieldTaggedSummaries,
)
from wallaroo.wallaroo_ml_ops_api_client.models.series_summary import SeriesSummary
from wallaroo.wallaroo_ml_ops_api_client.models.series_summary_statistics import (
    SeriesSummaryStatistics,
)


class StaticBaseline(MLOpsStaticBaseline):
    """A Baseline from the server will always come in the form of a set of Summaries."""

    def summarize(self):
        # TODO: Expose summarize function like v1 to convert between baseline types.
        pass

    def _get_baseline_end(self) -> datetime:
        return self.static[-1]  # type: ignore[return-value]

    def _get_display_row(self) -> str:
        return """
        <tr><td>Baseline</td><td>TODO</td></tr>
        """


class SummaryBaseline(MLOpsSummaryBaseline):
    """This class represents a summary baseline.
    A Baseline from the MLOps API will always come in the form of a set of Summaries.
    """

    def _get_display_row(self) -> str:
        # TODO: Check dates
        return f"""
        <tr><td>Monitoring</td><td>{list(self.summary.to_dict().keys())}</td></tr>
        """

    def _get_baseline_end(self) -> datetime:
        return self.summary[list(self.summary.to_dict())[0]].end  # type: ignore[return-value]

    @classmethod
    def _from_v1_config(cls, v1_config: V1AssayConfig) -> "SummaryBaseline":
        config_baseline = v1_config.baseline
        if isinstance(config_baseline, V1CalculatedBaseline) or isinstance(
            config_baseline, V1FixedBaseline
        ):
            start = cast(str, config_baseline.calculated["fixed_window"].get("start"))
            baseline_start_at = dateutil.parser.parse(start)
            end = cast(str, config_baseline.calculated["fixed_window"].get("end"))
            baseline_end_at = dateutil.parser.parse(end)
            return StaticBaseline([baseline_start_at, baseline_end_at]).summarize()
        elif isinstance(config_baseline, V1StaticBaseline):
            config_summarizer = cast(V1Summarizer, v1_config.summarizer)
            parsed_path = unwrap(v1_config.window.path).split()
            prefix = "in" if parsed_path[0] == "input" else "out"
            iopath = f"{prefix}.{parsed_path[1]}.{parsed_path[2]}"

            start = cast(str, config_baseline.static.get("start"))
            baseline_start_at = dateutil.parser.parse(start)
            end = cast(str, config_baseline.static.get("end"))
            baseline_end_at = dateutil.parser.parse(end)
            return SummaryBaseline.from_v1_summary(
                config_baseline, config_summarizer, iopath
            )
        else:
            raise Exception(
                "Could not parse baseline, unknown V1 type.",
                isinstance(config_baseline, V1StaticBaseline),
                config_baseline.__dict__,
            )

    @classmethod
    def from_v1_summary(
        cls, baseline: V1StaticBaseline, summarizer: V1Summarizer, iopath: str
    ) -> "SummaryBaseline":
        """Create a SummaryBaseline from a V1Summarizer.
        A v1 summary is guaranteed to only contain one observed path.

        :param baseline: The v1 baseline.
        :param summarizer: The v1 summarizer.
        :param iopath: The iopath.

        :return: The SummaryBaseline object.
        """

        aggregation = Aggregation(baseline.static["aggregation"].value)  # type: ignore # mypy can't handle complex dicts
        # TODO: Move baseline calculation to assays v2, this is relying on v1.
        v1_edges = cast(List[Union[float, str]], baseline.static.get("edges"))
        # TODO: Handle infinity, make sure edge cases are clean.
        if v1_edges[-1] is None:
            # We could convert None instead of a str to INF if this is weird.
            v1_edges[-1] = "INFINITY"
        v1_labels = cast(List[str], baseline.static.get("edge_names"))
        v1_mode = summarizer.bin_mode
        v1_bin_count = summarizer.num_bins

        # OpenAPI doesn't know that "INFINITY" is allowed. This is manually deserialized into a float::INF by serde
        # In practice we never construct these manually, but we're converting v1 Baselines into the v2 Baseline, which is
        # not really a public interface.
        v1_edges_hack = cast(List[float], list(v1_edges))

        # TODO: Handle add_explicit_edges
        v2_mode = cast(
            Union[BinModeType0, BinModeType1, BinModeType2, BinModeType4],
            BinModeType0.NONE,
        )
        if v1_mode == V1BinMode.NONE:
            v2_mode = BinModeType0.NONE
        elif v1_mode == V1BinMode.EQUAL:
            v2_mode = BinModeType1(v1_bin_count)
        elif v1_mode == V1BinMode.QUANTILE:
            v2_mode = BinModeType2(v1_bin_count)
        elif v1_mode == V1BinMode.PROVIDED:
            v2_mode = BinModeType4(v1_edges_hack)

        v2_bins = Bins(edges=v1_edges_hack, labels=v1_labels, mode=v2_mode)
        summ = SeriesSummary(
            aggregated_values=cast(
                List[float], baseline.static.get("aggregated_values")
            ),
            aggregation=aggregation,
            bins=v2_bins,
            name=iopath,
            statistics=SeriesSummaryStatistics(
                count=cast(int, baseline.static.get("count")),
                max_=cast(float, baseline.static.get("max")),
                mean=cast(float, baseline.static.get("mean")),
                median=cast(float, baseline.static.get("median")),
                min_=cast(float, baseline.static.get("min")),
                std=cast(float, baseline.static.get("std")),
            ),
            start=cast(Union[None, datetime], baseline.static.get("start", None)),
            end=cast(Union[None, datetime], baseline.static.get("end", None)),
        )
        summary = FieldTaggedSummaries.from_dict({iopath: summ.to_dict()})
        return SummaryBaseline(summary=summary)
